local tracker = ...
table.insert(Event.Addon.Load.End, {function(addon) if (addon == tracker.identifier) then myUI:RegisterModule(tracker) end end, tracker.identifier, "_startup"})
--Command.System.Strict()

local L = tracker.data.language

local ifnil = myUI.Utility.ifnil
local isnil = myUI.Utility.isnil
local rgba = myUI.Utility.rgba
local iif = myUI.Utility.iif
local pprint = myUI.Utility.pprint
local round = myUI.Utility.round
local minmax = myUI.Utility.minmax
local lpad = myUI.Utility.lpad
local trim = myUI.Utility.trim
local isempty = myUI.Utility.isempty
local ifempty = myUI.Utility.ifempty
local nilstr = myUI.Utility.nilstr
local notnil = myUI.Utility.notnil

local floor = math.floor
local abs = math.abs
local modf = math.modf

local function frac(n) local _, f = modf(n); return f end

local tinsert = table.insert
local tremove = table.remove
local len = string.len

local myuicolors = myUI.Colors
-- addon --

local _aprofile, _cprofile -- account and character profiles, respectively
local _configFrame = nil
local _sensitivity = 10	-- moving/resizing

local LAYER_MOVER = myUI.Layers.UnitFrame + 2

local _mover
local _boxes = {}
local _recycled = {}
local _tracking


local function _createMover()	-- mover creation code
	if (not _mover) then
		_mover = myUI:CreateFrame("Frame")
		
		_mover:SetBackgroundColor(rgba(myuicolors.Mover))
		_mover:SetLayer(LAYER_MOVER)
		_mover:SetVisible(false)
		_mover:SetSize(64)
		
		_mover.titletext = myUI:CreateFrame("Text", _mover)
		_mover.titletext:SetBackgroundColor(0.6,0,0,0.6)
		_mover.titletext:SetText(L["bufftracker"])
		_mover.titletext:SetPoint("BOTTOMLEFT", _mover, "TOPLEFT", 0, -4)
		_mover.titletext:SetFontSize(12); _mover.titletext.customFontSize = true
		
		_mover.coordtext = myUI:CreateFrame("Text", _mover)
		_mover.coordtext:SetBackgroundColor(0.6,0,0,0.6)
		_mover.coordtext:SetPoint("TOPLEFT", _mover.titletext, "TOPRIGHT", 4, 0)
		_mover.coordtext:SetFontSize(12); _mover.coordtext.customFontSize = true
		
		_mover.sensitivity = myUI:CreateFrame("Text", _mover)
		_mover.sensitivity:SetBackgroundColor(0.6,0,0,0.6)
		_mover.sensitivity:SetPoint("TOPLEFT", _mover.coordtext, "TOPRIGHT", 4, 0)
		_mover.sensitivity:SetFontSize(12); _mover.sensitivity.customFontSize = true
		_mover.sensitivity:SetText("+/- " .. tostring(_sensitivity))
		
		_mover:SetPoint("TOPLEFT", UIParent, "TOPLEFT", _aprofile.x, _aprofile.y)
		

		_mover.Show = function(mover)
			mover.coordtext:SetText(tostring(floor(mover:GetLeft())) .. ", " .. tostring(floor(mover:GetTop())))
			mover:SetVisible(true)
		end
		_mover.Hide = function(mover)
			mover:SetVisible(false)
		end
		
		_mover.Event.RightDown = function(mover)
			if (mover:GetVisible()) then
				local s = _sensitivity - 1
				if (s < 1) then s = 10 end
				
				_sensitivity = s
				mover.sensitivity:SetText("+/- " .. tostring(_sensitivity))
			end		
		end
		
		_mover.Event.LeftDown = function(mover)
			local m = Inspect.Mouse()
			if (m) then
				mover.x = m.x
				mover.y = m.y
				mover.dragging = true
			end
		end
		_mover.Event.LeftUp = function(mover)
			mover.dragging = false
		end
		_mover.Event.LeftUpoutside = function(mover)
			mover.dragging = false
		end
		_mover.Event.MouseMove = function(mover, x, y)
			if (mover.dragging) then
				local deltax = floor(((x - mover.x) / _sensitivity) * _sensitivity)
				local deltay = floor(((y - mover.y) / _sensitivity) * _sensitivity)
				
				if (abs(deltax) > 0) or (abs(deltay) > 0) then
					local xnow = mover:GetLeft() + deltax
					local ynow = mover:GetTop() + deltay
					
					mover:SetPoint("TOPLEFT", UIParent, "TOPLEFT", xnow, ynow)
										
					_aprofile.x = floor(xnow)
					_aprofile.y = floor(ynow)
					
					mover.coordtext:SetText(tostring(_aprofile.x) .. ", " .. tostring(_aprofile.y))
					
					mover.x = mover.x + deltax
					mover.y = mover.y + deltay				
				end			
			end
		end
	end
end



local LAYER_TRACKER = myUI.Layers.UnitFrame + 2
local LAYER_MOVER = LAYER_TRACKER + 20

local _onBuffAdd

local function _layoutBOX(box)
	box._blink = false
	if (box._buffid) then
		box:SetSize(_aprofile.trackingCXY)
		box:SetAlpha(_aprofile.trackingAlpha)
	else
		box:SetSize(_aprofile.warningCXY)
		box:SetAlpha(_aprofile.warningAlpha)
	end
end
local function _setBuffBOX(box, buffid)
	if (box._buffid ~= buffid) then
		box._blink = false
		
		box._buffid = buffid
		box:Layout()
	end
	box:SetVisible(true)
end
local function _getBuffBOX(box)
	return box._buffid
end

local function _disposeBOX(box)
	box._disposed = true
	box._buffid = nil
	box:SetVisible(false)
	tinsert(_recycled, box)
end

local function _newBOX(detail)
	local box
	if (#_recycled > 0) then
		box = tremove(_recycled)
		box._disposed = false
	else
		box = myUI:CreateFrame("Frame")
		box:SetVisible(false)
		box:SetBackgroundColor(1,1,1,1)
		box.icon = myUI:CreateFrame("Texture", box)
		box.icon:SetPoint("TOPLEFT", box, "TOPLEFT", 2, 2)
		box.icon:SetPoint("BOTTOMRIGHT", box, "BOTTOMRIGHT", -2, -2)
		
		box.Dispose = _disposeBOX
		box.Layout = _layoutBOX
		box.SetBuff = _setBuffBOX
		box.GetBuff = _getBuffBOX
	end

	box.icon:SetTexture("Rift", detail.icon)
	return box
end

local function _animationUpdate(self, now, elapsed)
	local wa = _aprofile.warningAlpha
	
	for _, box in pairs(_boxes) do
		if (not box._disposed) and (not box:GetBuff()) then
			-- warning state
	
			local a = box:GetAlpha()
			if (box._blink) then
				a = a + 0.1
			else
				a = a - 0.1
			end
			
			box:SetAlpha(minmax(a, 0.1, wa))
			if (a >= wa) then
				box._blink = false
			elseif (a <= 0.1) then
				box._blink = true
			end
		else
			box._blink = false
		end
	end
end



local function _layout(doBoxes)
	if (doBoxes) then
		for _, box in pairs(_boxes) do box:Layout() end
	end
	
	local buffnames = {}
	for buff, _ in pairs(_boxes) do
		tinsert(buffnames, buff)
	end
	
	local nboxes = #buffnames
	
	if (nboxes > 1) then table.sort(buffnames) end	-- so the order doesn't jump around
	if (nboxes > 0) then
		for _, box in pairs(_boxes) do 
			box:ClearPoint("CENTER")
			box:ClearPoint("LEFTCENTER")
			box:ClearPoint("RIGHTCENTER")
			box:SetVisible(true)
		end
		
		
		
		_boxes[buffnames[1]]:SetPoint("LEFTCENTER", _mover, "LEFTCENTER")
		
		
		for i=2, nboxes do
			_boxes[buffnames[i]]:SetPoint("LEFTCENTER", _boxes[buffnames[i-1]], "RIGHTCENTER", 4, 0)			
		end
	end		
end







local function _buildTrackingList(abilities, reset)
	if (reset or (not _tracking)) then
		if (not abilities) then 
			abilities = Inspect.Ability.New.List() 
		end
		
		if (abilities) then
			_tracking = {}
			
			if (reset) then
				for _, box in pairs(_boxes) do
					box:Dispose()
				end
				_boxes = {}
			end
			
			local details = Inspect.Ability.New.Detail(abilities)
			for abilityID, detail in pairs(details) do
				if (_cprofile.abilities[detail.name]) then
					for _, buffname in ipairs(_cprofile.abilities[detail.name]) do
						if (not _tracking[buffname]) then
							_tracking[buffname] = true					
						end				
					end
				end		
			end	
			
			if (reset) then
				_onBuffAdd(Inspect.Unit.Lookup("player"), Inspect.Buff.List("player"))
			end
		end	
	end
	return (_tracking ~= nil)
end

_onBuffAdd = function(unit, buffs)
	if (unit == Inspect.Unit.Lookup("player")) then
		if (_buildTrackingList()) then
			local details = Inspect.Buff.Detail("player", buffs)
			local dolayout
			
			if (details) then
				for buffid, detail in pairs(details) do
					if (_tracking[detail.name]) then
						-- we should be tracking this one
						if (not _boxes[detail.name]) then
							_boxes[detail.name] = _newBOX(detail)
							dolayout = true
						end
						if (not _boxes[detail.name]:GetBuff()) then
							_boxes[detail.name]:SetBuff(buffid)
						end
					
					end
				end
			end
			
			if (dolayout) then _layout() end
		end
	end
end

local function _onBuffRemove(unit, buffs)
	if (unit == Inspect.Unit.Lookup("player")) then
		if (_buildTrackingList()) then
			local bid 
			for _, box in pairs(_boxes) do
				bid = box:GetBuff()
				if (bid and buffs[bid]) then
					box:SetBuff(nil)	-- we're tracking and buff isn't present
				end
			end
		end		
	end
end




local function _onUnitAdd(units)
	if (not _tracking) then
		_buildTrackingList()
	end
end

local function _onAbilityAdd(abilities)
	_buildTrackingList(_, true)
end

local function _onAbilityRemove(abilities)
	_buildTrackingList(_, true)
end

local function _reset()
	_buildTrackingList(_, true)
end

-- public interface methods

function tracker:Initialize()
	local id = tracker.identifier
	
	
end

function tracker:LoadProfile(account, character)
	_aprofile = account
	_cprofile = character

	local adefs = tracker._getAccountDefaults()
	
	myUI.Utility.nilCopy(_aprofile, adefs)	
	if (not _cprofile.abilities) then _cprofile.abilities = {} end
	
	
	
	_createMover()
	
	local id = tracker.identifier
	
	myUI:HookEvent(Event.Buff.Add, _onBuffAdd, tracker, id .. "_onBuffAdd")
	myUI:HookEvent(Event.Buff.Remove, _onBuffRemove, tracker, id .. "_onBuffRemove")
	myUI:HookEvent(Event.Ability.New.Add, _onAbilityAdd, tracker, id .. "_onAbilityAdd")
	myUI:HookEvent(Event.Ability.New.Remove, _onAbilityRemove, tracker, id .. "_onAbilityRemove")
	myUI:HookEvent(Event.Unit.Add, _onUnitAdd, tracker, id .. "_onUnitAdd")			-- fires on startup
	
	
	myUI:HookEvent(Event.Unit.Availability.Full, _onUnitAdd, tracker, id .. "_onUnitAdd")	-- fires on reload
	
	if (not myUI.Events.AnimationUpdate:IsSubscribed(id .. "._animationUpdate")) then
		myUI.Events.AnimationUpdate:Subscribe(id .. "._animationUpdate", tracker, _animationUpdate)
	end
	
	if (_configFrame) then _configFrame:LoadProfile() end	
end

function tracker:GetConfigurationFrame(parent)
	if (not _configFrame) then
		Command.System.Watchdog.Quiet()
		_configFrame = myUI:CreateFrame("Frame", parent)	
		
		local vp = myUI:CreateFrame("myUI_Viewport", _configFrame)
		vp:SetPoint("TOPLEFT", _configFrame, "TOPLEFT", 0, 0)
		vp:SetPoint("BOTTOMRIGHT", _configFrame, "BOTTOMRIGHT", 0, 0)
		
		local content = vp.Content
		content:SetHeight(700)
		--content:SetBackgroundColor(0.6,0,0,0.2)
		
		local margin = 16
		local buffList, abilityList
		local txtAddAbility, txtAddBuff
		local btnDeleteAbility, btnDeleteBuff
		
		-- normal controls
		local hdr = myUI:CreateFrame("Text", content)
		hdr:SetText(L["tracking"])
		hdr:SetPoint("TOPLEFT", content, "TOPLEFT", margin, margin)
		
		local trackingAlpha = myUI:CreateFrame("RiftSlider", content)
		trackingAlpha:SetPoint("TOPLEFT", hdr, "BOTTOMLEFT", 120, 8)
		trackingAlpha:SetWidth(100)
		trackingAlpha:SetRange(0,10)
		
		trackingAlpha.Event.SliderChange = function(slider)
			if (not _configFrame._loading) then
				_aprofile.trackingAlpha = slider:GetPosition()/10
				_layout(true)
			end
		end
		
		local label = myUI:CreateFrame("Text", content)
		label:SetText(L["alpha"])
		label:SetPoint("RIGHTCENTER", trackingAlpha, "LEFTCENTER", -4, -8)
		
		local trackingCXY = myUI:CreateFrame("RiftSlider", content)
		trackingCXY:SetPoint("TOPLEFT", trackingAlpha, "BOTTOMLEFT")
		trackingCXY:SetWidth(100)
		trackingCXY:SetRange(12, 60)
		
		trackingCXY.Event.SliderChange = function(slider)
			if (not _configFrame._loading) then
				_aprofile.trackingCXY = slider:GetPosition()
				_layout(true)
			end
		end
		
		label = myUI:CreateFrame("Text", content)
		label:SetText(L["cxy"])
		label:SetPoint("RIGHTCENTER", trackingCXY, "LEFTCENTER", -4, -8)
		
		label = myUI:CreateFrame("Text", content)
		label:SetText(L["warning"])
		label:SetPoint("TOPLEFT", hdr, "TOPRIGHT", 300, 0)
		
		hdr = label
		
		local warningAlpha = myUI:CreateFrame("RiftSlider", content)
		warningAlpha:SetPoint("TOPLEFT", hdr, "BOTTOMLEFT", 120, 8)
		warningAlpha:SetWidth(100)
		warningAlpha:SetRange(2, 10)
		
		warningAlpha.Event.SliderChange = function(slider)
			if (not _configFrame._loading) then
				_aprofile.warningAlpha = slider:GetPosition() / 10
				_layout(true)
			end
		end
		
		label = myUI:CreateFrame("Text", content)
		label:SetText(L["alpha"])
		label:SetPoint("RIGHTCENTER", warningAlpha, "LEFTCENTER", -4, -8)
		
		local warningCXY = myUI:CreateFrame("RiftSlider", content)
		warningCXY:SetPoint("TOPLEFT", warningAlpha, "BOTTOMLEFT")
		warningCXY:SetWidth(100)
		warningCXY:SetRange(12, 120)
		
		warningCXY.Event.SliderChange = function(slider)
			if (not _configFrame._loading) then
				_aprofile.warningCXY = slider:GetPosition()
				_layout(true)
			end
		end
		
		label = myUI:CreateFrame("Text", content)
		label:SetText(L["cxy"])
		label:SetPoint("RIGHTCENTER", warningCXY, "LEFTCENTER", -4, -8)
		
		-- abilities lists
		local abilityList = myUI:CreateFrame("myUI_Listbox", content)
		abilityList:SetSize(240, 360)
		abilityList:SetPoint("TOPLEFT", content, "TOPLEFT", 64, 140)
		abilityList.Viewport:SetBackgroundColor(rgba(myuicolors.Selected))
		
		abilityList.Event.SelectedIndexChanged = function(list, selindexes)
			if (not _configFrame._loading) then
				buffList:ClearItems()
				
				if (selindexes and (#selindexes > 0)) then
					for _, buff in ipairs(_cprofile.abilities[list:GetItem(selindexes[1])]) do
						buffList:AddItem(buff)
					end
					
					btnDeleteAbility:SetText(L["delete"] .. " \"" .. list:GetItem(selindexes[1]) .. "\"")
					btnDeleteAbility:SetEnabled(true)
				else
					btnDeleteAbility:SetText(" ")
					btnDeleteAbility:SetEnabled(false)
				end
			end
		end
		
		label = myUI:CreateFrame("Text", content)
		label:SetText(L["abilities"])
		label:SetPoint("BOTTOMLEFT", abilityList, "TOPLEFT")
		
		txtAddAbility = myUI:CreateFrame("myUI_Textbox", content)
		txtAddAbility:SetWidth(160)
		txtAddAbility:SetPoint("TOPRIGHT", abilityList, "BOTTOMRIGHT", -20, 8)
		txtAddAbility:SetText("")
				
		txtAddAbility.Event.OnEnter = function(txt)
			local a = trim(txt:GetText())
			if (len(a) > 0) then
				local exists
				for ability, _ in pairs(_cprofile.abilities) do
					if (ability == a) then
						exists = true
						break
					end
				end
				if (not exists) then
					_cprofile.abilities[a] = {}
					txt:SetText("")
					abilityList:AddItem(a)	

					_reset()
				end
				
			end
		end
		
		btnDeleteAbility = myUI:CreateFrame("myUI_Button", content)
		btnDeleteAbility:SetText(" ")
		btnDeleteAbility:SetEnabled(false)
		btnDeleteAbility:SetPoint("TOPLEFT", txtAddAbility, "BOTTOMLEFT", 0, 8)
		btnDeleteAbility:SetPoint("TOPRIGHT", txtAddAbility, "BOTTOMRIGHT", 0, 8)
		
		btnDeleteAbility.Event.LeftPress = function(btn)			
			local inx = abilityList:GetSelectedIndexes()
			if (inx and (#inx > 0)) then
				buffList:ClearItems()
				local a = abilityList:GetItem(inx[1])
				
				_cprofile.abilities[a] = nil				
				abilityList:RemoveItemAt(inx[1])
				
				_reset()
			end
		end
		
		label = myUI:CreateFrame("Text", content)
		label:SetText(L["add"])
		label:SetPoint("RIGHTCENTER", txtAddAbility, "LEFTCENTER", -8, 0)
		
		buffList = myUI:CreateFrame("myUI_Listbox", content)
		buffList:SetSize(240, 360)
		buffList:SetPoint("TOPLEFT", abilityList, "TOPRIGHT", 100, 0)
		buffList.Viewport:SetBackgroundColor(rgba(myuicolors.Selected))
		
		buffList.Event.SelectedIndexChanged = function(list, selindexes)
			if (not _configFrame._loading) then
				if (selindexes and (#selindexes > 0)) then
					btnDeleteBuff:SetEnabled(true)
					btnDeleteBuff:SetText(L["delete"] .. " \"" .. list:GetItem(selindexes[1]) .. "\"")
				else
					btnDeleteBuff:SetEnabled(false)
					btnDeleteBuff:SetText(" ")
				end
			end
		end
		
		local label = myUI:CreateFrame("Text", content)
		label:SetPoint("BOTTOMLEFT", buffList, "TOPLEFT")
		label:SetText(L["buffs"])
		
		txtAddBuff = myUI:CreateFrame("myUI_Textbox", content)
		txtAddBuff:SetWidth(160)
		txtAddBuff:SetPoint("TOPRIGHT", buffList, "BOTTOMRIGHT", -20, 8)
		txtAddBuff:SetText("")
		
		txtAddBuff.Event.OnEnter = function(txt)
			local b = trim(txt:GetText())
			if (len(b) > 0) then
				local inx = abilityList:GetSelectedIndexes()
				if (inx and (#inx > 0)) then
					local a = abilityList:GetItem(inx[1])
					
					local exists
					for buff, _ in ipairs(_cprofile.abilities[a]) do
						if (buff == b) then
							exists = true
							break
						end
					end
					if (not exists) then
						tinsert(_cprofile.abilities[a], b)
						buffList:AddItem(b)
						txt:SetText("")
						
						_reset()
					end
				end
			end
		end
		
		label = myUI:CreateFrame("Text", content)
		label:SetText(L["add"])
		label:SetPoint("RIGHTCENTER", txtAddBuff, "LEFTCENTER", -8, 0)
		
		btnDeleteBuff = myUI:CreateFrame("myUI_Button", content)
		btnDeleteBuff:SetText(" ")
		btnDeleteBuff:SetEnabled(false)
		btnDeleteBuff:SetPoint("TOPLEFT", txtAddBuff, "BOTTOMLEFT", 0, 8)
		btnDeleteBuff:SetPoint("TOPRIGHT", txtAddBuff, "BOTTOMRIGHT", 0, 8)
		
		btnDeleteBuff.Event.LeftPress = function(btn)
			local inx = abilityList:GetSelectedIndexes()
			if (inx and (#inx > 0)) then
				a = abilityList:GetItem(inx[1])
				inx = buffList:GetSelectedIndexes()
				if (inx and (#inx > 0)) then
					b = buffList:GetItem(inx[1])
					
					local removeAt
					for i, buff in ipairs(_cprofile.abilities[a]) do
						if (buff == b) then
							removeAt = i
							break
						end					
					end
					if (removeAt) then
						tremove(_cprofile.abilities[a], removeAt)
						buffList:RemoveItemAt(inx[1])
						
						_reset()
					end
				end	
			end
		end
		
		_configFrame.LoadProfile = function(self)
			self._loading = true
			
			trackingAlpha:SetPosition(_aprofile.trackingAlpha * 10)
			trackingCXY:SetPosition(_aprofile.trackingCXY)
			warningAlpha:SetPosition(_aprofile.warningAlpha * 10)
			warningCXY:SetPosition(_aprofile.warningCXY)
		
			abilityList:ClearItems()
			buffList:ClearItems()
		
			for ability, _ in pairs(_cprofile.abilities) do
				abilityList:AddItem(ability)
			end
		
			self._loading = false
		end
		_configFrame:LoadProfile()
	end
	return _configFrame
end

function tracker:ShowMovers()
	_mover:Show()
		
	
end
function tracker:HideMovers()
	_mover:Hide()
		
	
end

--
